\section{Discovering Blocked Algorithms}


\begin{code}[hide]
open import Data.Nat as ℕ
open import Data.Nat.Properties as ℕ
open import Data.Fin as Fin using (Fin; zero; suc; remQuot; combine; quotRem)
open import Data.Fin.Properties as Fin using (remQuot-combine; combine-remQuot)
open import Relation.Binary.PropositionalEquality as Eq hiding ([_])
open import Relation.Binary
open import Data.Product as Prod hiding (map)
open import Data.Sum hiding (map; swap)
open import Function
open import Algebra
open import Level using () renaming (zero to ℓ₀)


variable
  X Y Z : Set
  m n k : ℕ

infixl 5 _⊗_
\end{code}

The initial point of our framework is a dependently-typed
language.  We use Agda, but none of the constructions presented
below are Agda-specific, so any implementation of Martin-L\"of
type theory can be used instead.  Our formalisation introduces
a minimalist rank-polymorphic array theory, derives
blocked matrix multiplication and proves its correctness.
This section is written in literate Agda, so all the
expressions have been typechecked.  We assume some
familiarity with Agda syntax, otherwise consider folllowing
any of these\footnote{The list of introductory tutorials to Agda
can be found here:
\url{https://agda.readthedocs.io/en/v2.6.3/getting-started/tutorial-list.html}}
tutorials.


When working with dependent types it is important to
chose the right encoding for the data that we are reasoning
about.  A well-chosen encoding can either make all further
reasoning pleasant or turn it into a nightmare.

In our case, we have to chose representation for multi-dimensional
arrays.  We want to be able to reason about rank-polymorphic
functions on these arrays and guarantee safe indexing.
We are not interested in capturing
low-level implementational details such as memory allocation
or synchronisation of threads.  For such cases, it is a common
practice to represent arrays as functions.  For the
purposes of rank-polymorphism, we have to chose a convenient
representation for array shapes and array indices.

Array shapes are sequences of natural numbers.  While these
can be represented as lists, we encode them as finite binary trees,
where leaves are natural numbers.  Such encoding makes it
possible to describe non-trivial traversal of the shape sequence
inductively.  We call the data type for shapes \AD{S}, and we
define it inductively using two constructors.
\begin{code}
data S : Set where
  ι    : ℕ → S
  _⊗_  : S → S → S
\end{code}
Leaves of the shape tree are constructed with \AC{ι} which takes
one argument.  The \AC{\_⊗\_} constructor creates a tree out of
two subtrees which are the arguments of the constructor.  Note
that underscores in \AC{\_⊗\_} specify the positions where the
arguments go, therefore \AC{⊗} is a binary operation.

\begin{code}[hide]
variable
  s p q r u v w : S
\end{code}

Array indices (positions) are given by the dependent type \AD{P}
which is indexed by array shapes.  A position for the given shape
has exactly the same tree structure as the shape, but all the leaves
are natural numbers that are bound by the shape leaf.  Again,
\AD{P} is given inductively using the following two constructors.
\begin{code}
data P : S → Set where
  ι    : Fin n → P (ι n)
  _⊗_  : P s → P p → P (s ⊗ p)
\end{code}

The type of arrays is indexed by the element type and the shape.
We use a C-like syntax with square brackets after the element
type.  Objects of this type are functions from array indices
into array elements.
\begin{code}
_⟪_⟫ : Set → S → Set
X ⟪ s ⟫ = P s → X
\end{code}

There is no need for special syntax for creating arrays, as
this can be achieved by a lambda term.  Neither there is a need
for the syntax for array selections, as this can be achieved with function
application.  Yet, for aesthetic purposes, we add a C-like
syntax for array selections:
\begin{code}
_[_]  : X ⟪ s ⟫ → (P s → X)
a [ ix ] = a ix
\end{code}
%syntax imap (λ i → e) = ⟪ i ⇒ e ⟫


\paragraph{Summation}
As a prerequisite for defining matrix multiplication, we have
to define array summation.  We define a 1-dimensional version
of summation called \AF{sum₁}.  After that, we will generalise
it to arbitrary shapes.

We start with a few helper operations: \AF{ιsuc} increases the
given 1-dimensional index by 1; \AF{hd} is the head (first element
of the 1-dimensional array); \AF{tl} is the tail (the 1-dimensional
array without the first element).
\begin{code}
ιsuc : P (ι n) → P (ι (suc n))
ιsuc (ι i) = ι (suc i)

hd : X ⟪ ι (suc n) ⟫ → X
hd a = a [ ι zero ]

tl : X ⟪ ι (suc n) ⟫ → X ⟪ ι n ⟫
tl a i = a [ ιsuc i ]
\end{code}

With these operations in place, \AF{sum₁} is defined recursively.
The function has two arguments: the binary function, the initial
element and the 1-dimensional array that we sum.
For the empty array we return the initial element, and for non-empty
arrays we add the head of the array to the sum of the tail.
\begin{code}
sum₁ : (X → X → X) → X → X ⟪ ι n ⟫ → X
sum₁ {n = zero}   f e a = e
sum₁ {n = suc n}  f e a = f (hd a) (sum₁ f e (tl a))
\end{code}


\paragraph{Array Combinators}
It is useful to invest a little time in defining common array combinators.
As our arrays are functions, we can easily define:
\AF{K} \AB{x} to produce an array where all the elements are $x$;
\AF{map} \AB{f} \AB{a} to apply $f$ to all the elements of $a$; and
\AF{zipWith} \AB{f} \AB{a} \AB{b} to apply the binary operation $f$ to
arrays $a$ and $b$ point-wise, given that $a$ and $b$ are of the same shape.
\begin{code}
K : X → X ⟪ s ⟫
K x i = x

map : (X → Y) → X ⟪ s ⟫ → Y ⟪ s ⟫
map f a i = f (a [ i ])

zipWith : (X → Y → Z) → X ⟪ s ⟫ → Y ⟪ s ⟫ → Z ⟪ s ⟫
zipWith f a b i = f (a [ i ]) (b [ i ])
\end{code}

The \AF{nest} and \AF{unnest} combinators make it possible to treat the same
array as a nested one (array of arrays); or as an array of a product shape.
Note that both operations change the order of the shape components because
our array notation (\AF{\_⟪\_⟫}) is a postfix operation.
\begin{code}
nest : X ⟪ s ⊗ p ⟫ → X ⟪ p ⟫ ⟪ s ⟫
nest a i j = a [ i ⊗ j ]

unnest : X ⟪ p ⟫ ⟪ s ⟫ → X ⟪ s ⊗ p ⟫
unnest a (i ⊗ j) = a [ i ] [ j ]
\end{code}



\paragraph{Generalised Sum}
We define \AF{sum} for arbitrarily shaped array by induction
on the shape.  For 1-d case, we use \AF{sum₁}.
For product shapes we apply \AF{sum} to all
the sub-arrays and sum all the partial results.
\begin{code}
sum : (X → X → X) → X → X ⟪ s ⟫ → X
sum {s = ι x}    f e a = sum₁  f e a
sum {s = s ⊗ p}  f e a = sum   f e $ map (sum f e) $ nest a
\end{code}
Note that this definition of \AF{sum} traverses the array
bottom-up, and it adds the neutral element after each
1-dimensional traversal.  We do not impose any
restrictions on the binary operation or the initial element
in this definition.


% FIXME think whether it is worth keeping this
% The binary operation does not need to be associative and
% the neutral element does not have to act like left or right
% zero.  By maintaining such a generality, we will see that
% correctness of blocking does not rely on this properties.
% Therefore, we can block a generalised matrix multiplication
% without any assumptions on addition, multiplication or the
% neutral element.

\paragraph{Canonical Matrix Multiplication}
Having the \AF{sum} operation at hand, we define
a canonical version of matrix multiplication.  In order to
keep matrix multiplication generic, we parametrise
it by the types $X$, $Y$ and $Z$ and
the operations (\AF{\_⊠\_} : \AB{X} → \AB{Y} → \AB{Z}),
(\AF{ε} : Z) and (\AF{\_⊞\_ : \AB{Z} → \AB{Z} → \AB{Z}})
that correspond to multiplication, initial element and
addition.

\begin{code}[hide]
module MMnaive (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z)
               (let infixl 10 _⊠_; _⊠_ = _⊠_) where
\end{code}
\begin{code}
  mm-canon : X ⟪ s ⊗ p ⟫ → Y ⟪ p ⊗ q ⟫ → Z ⟪ s ⊗ q ⟫
  mm-canon a b (i ⊗ j) =
    sum _⊞_ ε λ k → a [ i ⊗ k ] ⊠ b [ k ⊗ j ]
\end{code}
For now, there is no assumptions about the properties of
\AF{\_⊠\_}, \AF{ε} or \AF{\_⊞\_}.  Note that \AF{mm-canon}
is not restricted to two-dimensional arrays where the shape
is a product of two singletons (\eg{} \AC{ι} \AB{m} \AF{⊗}
\AC{ι} {n}), it is rank-polymorphic after all.  We come back
to this question later in this section.


\paragraph{Blocking}
The essence of blocked algorithms comes from the ability
to ``block'' arrays into smaller sub-arrays.  That is, we
cut the matrix into sub-matrices that preserve all the local
neighbours within the block.  For example,
consider an array of shape $4\times 6$ blocked into
an array of shape $2 \times 2$ where each element is a
$2 \times 3$ block.
\[
\begin{pmatrix}
a & a & a & a & a & a \\
b & b & b & b & b & b \\
c & c & c & c & c & c \\
d & d & d & d & d & d
\end{pmatrix}
\Rightarrow
\left(
\begin{array}{c c c | c c c}
a & a & a & a & a & a \\
b & b & b & b & b & b \\
\hline
c & c & c & c & c & c \\
d & d & d & d & d & d
\end{array}
\right)
\]
For our arrays, we can define conversion functions that perform
blocking over rows and columns as presented above.
\begin{code}
block : X ⟪ (s ⊗ p) ⊗ (q ⊗ r) ⟫ → X ⟪ (s ⊗ q) ⊗ (p ⊗ r) ⟫
block a ((i ⊗ j) ⊗ (k ⊗ l)) = a [ (i ⊗ k) ⊗ (j ⊗ l) ]

unblock : X ⟪ (s ⊗ q) ⊗ (p ⊗ r) ⟫ → X ⟪ (s ⊗ p) ⊗ (q ⊗ r) ⟫
unblock = block
\end{code}
Note that \AF{block} is a self-inverse, but we will will define
the \AF{unblock} to indicate the intention.

Next, we can make a cut over the columns of the matrix.  For
the example above of shape $4\times 6$, we get an array of shape
$3$ where the elements are blocks of shape $4 \times 2$
\[
\begin{pmatrix}
a & a & a & a & a & a \\
b & b & b & b & b & b \\
c & c & c & c & c & c \\
d & d & d & d & d & d
\end{pmatrix}
\Rightarrow
\left(
\begin{array}{c c | c c | c c}
a & a & a & a & a & a \\
b & b & b & b & b & b \\
c & c & c & c & c & c \\
d & d & d & d & d & d
\end{array}
\right)
\]
We call this operation \AF{blockᵥ} (vertical cuts), and similarly
to \AF{block} it is a self-inverse, but we still introduce
\AF{unblockᵥ} to indicate the intention.
\begin{code}
blockᵥ : X ⟪ s ⊗ (p ⊗ q) ⟫ → X ⟪ p ⊗ (s ⊗ q) ⟫
blockᵥ a (i ⊗ (j ⊗ k)) = a [ j ⊗ (i ⊗ k) ]

unblockᵥ : X ⟪ p ⊗ (s ⊗ q) ⟫ → X ⟪ s ⊗ (p ⊗ q) ⟫
unblockᵥ = blockᵥ
\end{code}


Finally, we can block over rows (horizontal cuts), and for our $4 \times 6$
example we have an array of $2$ where elements are
blocks of shape $2 \times 6$.
\[
\begin{pmatrix}
a & a & a & a & a & a \\
b & b & b & b & b & b \\
c & c & c & c & c & c \\
d & d & d & d & d & d
\end{pmatrix}
\Rightarrow
\left(
\begin{array}{c c c c c c}
a & a & a & a & a & a \\
b & b & b & b & b & b \\
\hline
c & c & c & c & c & c \\
d & d & d & d & d & d
\end{array}
\right)
\]
In this case, the function is not a self-inverse
as we have to rearrange the structure of the shape
\begin{code}
blockₕ : X ⟪ (s ⊗ p) ⊗ q ⟫ → X ⟪ s ⊗ (p ⊗ q) ⟫
blockₕ a (i ⊗ (j ⊗ k)) = a [ (i ⊗ j) ⊗ k ]

unblockₕ : X ⟪ s ⊗ (p ⊗ q) ⟫ → X ⟪ (s ⊗ p) ⊗ q ⟫
unblockₕ a ((i ⊗ j) ⊗ k) = a [ i ⊗ (j ⊗ k) ]
\end{code}

Note that while our pictures are two- and four-dimensional,
the blocking functions are rank-polymorphic.  Free variables
\AB{s}, \AB{p}, \AB{q} and \AB{r} can be of arbitrary ranks.

\paragraph{Blocked Matrix Multiplication}
We are ready to define blocked matrix multiplication that we call
\AF{mm} and that has the following type for some shapes $s$, $p$ and
$q$:
\begin{code}[hide]
module MMblocked (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z)
         (let infixl 10 _⊠_; _⊠_ = _⊠_)
         (let infixl 5 _⊞_; _⊞_ = _⊞_) where

  open MMnaive _⊠_ ε _⊞_
  ni : (A : Set) → A → A
  ni A a = a

  infix -2 ni
  syntax ni A e = e ∶ A

\end{code}
\begin{code}
  mm : X ⟪ s ⊗ p ⟫ → Y ⟪ p ⊗ q ⟫ → Z ⟪ s ⊗ q ⟫
\end{code}

Now we define \AF{mm} by pattern-match on the structure of 
$s$, $p$ and $q$.  This gives us $2\cdot 2\cdot 2 = 8$ variants
that are given below.  The base-case of
our algorithm is the variant where $s$, $p$ and $q$ are
all singletons, in which case we simply apply the canonical
matrix multiplication.
\begin{code}
  mm {s = ι m} {ι n} {ι k} = mm-canon
\end{code}

In all the other cases, we block the input matrices in
such a way that the shapes of the blocks are appropriate
for the subsequent \AF{mm} calls.  Then we apply \AF{mm}
to these blocks and reduce partially-multiplied blocks using
\AF{sum} that is adjusted to add arbitrarily-shaped arrays.
We use the colon (\AF{∶}) notation to annotate the type
of the reblocked arrays before applying \AF{mm} recursively.
Finally, we have to unblock the result that we get from
the recursive call.
\begin{code}
  mm {s = ι m} {ι n} {p ⊗ q} a b =
    let  b′ = nest $ blockᵥ b                   ∶ Y ⟪ ι n ⊗ q ⟫ ⟪ p ⟫
         c′ = map (mm a) b′                     ∶ Z ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
    in   unblockᵥ $ unnest c′

  mm {s = ι m} {p ⊗ q} {ι n} a b =
    let  a′ = nest $ blockᵥ a                   ∶ X ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
         b′ = nest $ blockₕ b                   ∶ Y ⟪ q ⊗ ι n ⟫ ⟪ p ⟫
    in   sum (zipWith _⊞_) (K ε) λ k → mm (a′ [ k ]) (b′ [ k ])

  mm {s = ι m} {p ⊗ q} {r ⊗ w} a b =
    let  a′ = nest $ blockᵥ a                   ∶ X ⟪ ι m ⊗ q ⟫ ⟪ p ⟫
         b′ = nest $ block b                    ∶ Y ⟪ q ⊗ w ⟫ ⟪ p ⊗ r ⟫
         c′ = λ (i : P r) → sum (zipWith _⊞_) (K ε)
                                λ k → mm (a′ [ k ]) (b′ [ k ⊗ i ])
    in   unblockᵥ $ unnest c′


  mm {s = s ⊗ p} {ι m} {ι n} a b =
    let  a′ = nest $ blockₕ a                   ∶ X ⟪ p ⊗ ι m ⟫ ⟪ s ⟫
         c′ = map (flip mm b) a′                ∶ Z ⟪ p ⊗ ι n ⟫ ⟪ s ⟫
    in   unblockₕ $ unnest c′

  mm {s = s ⊗ p}  {ι m} {q ⊗ r} a b =
    let  a′ = nest $ blockₕ a                   ∶ X ⟪ p ⊗ ι m ⟫ ⟪ s ⟫
         b′ = nest $ blockᵥ b                   ∶ Y ⟪ ι m ⊗ r ⟫ ⟪ q ⟫
         c′ = (λ { (i ⊗ j) → mm (a′ [ i ])
                                (b′ [ j ]) })   ∶ Z ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
    in   unblock $ unnest c′


  mm {s = s ⊗ p} {q ⊗ r} {ι m} a b =
    let  a′ = nest $ block a                    ∶ X ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
         b′ = nest $ blockₕ b                   ∶ Y ⟪ r ⊗ ι m ⟫ ⟪ q ⟫
         c′ = λ (i : P s) → sum (zipWith _⊞_) (K ε)
                                λ k → mm (a′ [ i ⊗ k ]) (b′ [ k ])
    in   unblockₕ $ unnest c′

  mm {s = s ⊗ p} {q ⊗ r} {u ⊗ w} a b =
    let  a′ = nest $ block a                    ∶ X ⟪ p ⊗ r ⟫ ⟪ s ⊗ q ⟫
         b′ = nest $ block b                    ∶ Y ⟪ r ⊗ w ⟫ ⟪ q ⊗ u ⟫
         c′ = λ { (i ⊗ j) → sum (zipWith _⊞_) (K ε)
                                λ k → mm (a′ [ i ⊗ k ])
                                         (b′ [ k ⊗ j ]) }
    in   unblock $ unnest c′
\end{code}
It is satisfyingly surprising that all these cases appear naturally
by following the structure of the argument shapes.


\paragraph{Correctness}
While there is a strong evidence that blocked \AF{mm} implements matrix
multiplication, we would like to prove this formally. We introduce an
equivalence relation on arrays, which is defined as point-wise equality of the
array elements.
\begin{code}
_≈ₐ_ : X ⟪ s ⟫ → X ⟪ s ⟫ → Set
a ≈ₐ b = ∀ i → a [ i ] ≡ b [ i ]
\end{code}
% We can show congruence and other properties.
\begin{code}[hide]
refl-≈ₐ : {a : X ⟪ s ⟫} → a ≈ₐ a
refl-≈ₐ ix = refl

hd-cong : ∀ {n}{a b : X ⟪ ι (suc n) ⟫}
        → a ≈ₐ b → hd a ≡ hd b
hd-cong a≈b = a≈b (ι zero)

tl-cong : ∀ {n}{a b : X ⟪ ι (suc n) ⟫}
        → a ≈ₐ b → tl a ≈ₐ tl b
tl-cong a≈b i = a≈b (ιsuc i)

tl-comm : ∀ {n}{a : X ⟪ p ⟫ ⟪ ι (suc n) ⟫}
        → ∀ i → tl (λ j → a j i) ≡ λ j → tl a j i
tl-comm {n = n} i = refl

sum₁-comm : ∀ {n}{a : X ⟪ p ⟫ ⟪ ι n ⟫ }{f : X → X → X}{e : X}
          → sum₁ (zipWith f) (λ _ → e) a
            ≈ₐ λ i → sum₁ f e λ j → a j i
sum₁-comm {n = zero}  {a} {f} i = refl
sum₁-comm {n = suc n} {a} {f} i = cong₂ f refl (sum₁-comm {a = tl a} i)


sum₁-cong : ∀ {n}{a b : X ⟪ ι n ⟫}{f : X → X → X}{e : X}
          → a ≈ₐ b → sum₁ f e a ≡ sum₁ f e b
sum₁-cong {n = zero} a≈b = refl
sum₁-cong {n = suc n}{a = a}{b}{f}{e} a≈b = cong₂ f (hd-cong a≈b) (sum₁-cong $ tl-cong a≈b)

sum-cong : ∀ {s}{a b : X ⟪ s ⟫}{f : X → X → X}{e : X}
         → a ≈ₐ b → sum f e a ≡ sum f e b
sum-cong {s = ι x} a≈b = sum₁-cong a≈b
sum-cong {s = s ⊗ p} a≈b = sum-cong {s = s} λ i → sum-cong {s = p} λ j → a≈b _

sum-comm′ : {a : X ⟪ s ⊗ p ⟫}{f : X → X → X}{e : X}
          → sum (zipWith f) (λ _ → e) (nest a)
            ≈ₐ λ i → sum f e λ j → a (j ⊗ i)
sum-comm′ {s = ι x} i = sum₁-comm {n = x} i
sum-comm′ {s = s ⊗ p}{p = p₁}{a = a}{f}{e} i =
  let
    a′ = unnest (λ x →
              sum {s = p} (zipWith f) (λ _ → e)
              (λ j j₁ → a ((x ⊗ j) ⊗ j₁)))
    t = sum-comm′ {a = a′}{f}{e} i
  in trans t (sum-cong {s = s} λ ix → 
       let r = sum-comm′ { a = unnest (λ j j₁ → a (ix ⊗ j ⊗ j₁)) }{f}{e} i in r)


sum-comm : {a : X ⟪ p ⟫ ⟪ s ⟫}{f : X → X → X}{e : X}
          → sum (zipWith f) (K e) a
            ≈ₐ λ i → sum f e λ j → a [ j ] [ i ]
sum-comm {a = a} = sum-comm′ {a = unnest a}
\end{code}



We can formulate and prove our main theorem called
\AF{blocked-ok} that says: for any two matrices $a$ and $b$
(of the matching shapes) \AF{mm} $a$ $b$ computes the same
array as \AF{mm-canon} $a$ $b$.
\begin{code}[hide]
module Proof (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z) where

  open MMnaive _⊠_ ε _⊞_
  open MMblocked _⊠_ ε _⊞_
\end{code}

\begin{code}
  blocked-ok : (a : X ⟪ s ⊗ p ⟫) → (b : Y ⟪ p ⊗ q ⟫)
             → mm a b ≈ₐ mm-canon a b
\end{code}
For space purposes, we omit the proof itself, but it
can be found in the sources of this paper~\cite{paper-src}.
The proof itself is straight-forward, as the representation of
our arrays have very good computational properties.  Term normalisation
inlines most of the operations, and typically, all we have to do
is to apply the following lemma:
\begin{code}
  sum-commutes : (a : Z ⟪ p ⟫ ⟪ s ⟫)
               → sum (zipWith _⊞_) (K ε) a
                 ≈ₐ λ i → sum _⊞_ ε λ j → a [ j ] [ i ]
\end{code}
It says that for the nested array \AB{a}, it does not matter whether we
apply \AF{sum} with the operation lifted by \AF{zipWith} and the inital
element lifted by \AF{K}, or we apply the non-lifted \AF{sum} to all
the ``columns'' along the $p$ shape.  The proof of this lemma follows the definition
of \AF{sum}.

From this proof we see that blocking does not
depend on the properties of the operations used in matrix multiplication.
If the definition of the \AF{sum} is fixed, blocking does not change
the order in which the elements are multiplied and added.

\begin{code}[hide]
  sum-commutes a = sum-comm {a = a}

  blocked-ok {s = ι m} {ι n} {ι k} a b
    = refl-≈ₐ
  blocked-ok {s = ι m} {ι n} {p ⊗ q} a b (i ⊗ (j′ ⊗ k′))
    = blocked-ok a (blockᵥ b ∘ (j′ ⊗_)) (i ⊗ k′)

  blocked-ok {s = ι m} {p ⊗ q} {ι n} a b (i′ ⊗ j′)
    = let
        t k = mm (blockᵥ a ∘ (k ⊗_)) (blockₕ b ∘ (k ⊗_))
        r k = blocked-ok (blockᵥ a ∘ (k ⊗_)) (blockₕ b ∘ (k ⊗_)) (i′ ⊗ j′)
      in trans (sum-comm {a = t} (i′ ⊗ j′)) (sum-cong r)

  blocked-ok {s = ι m} {p ⊗ q} {r ⊗ w} a b (i′ ⊗ (j′ ⊗ k′))
    = let
        t k = mm (blockᵥ a ∘ (k ⊗_)) (block b ∘ (k ⊗ j′ ⊗_))
        r k = blocked-ok (blockᵥ a ∘ (k ⊗_)) (block b ∘ (k ⊗ j′ ⊗_)) (i′ ⊗ k′)
      in trans (sum-comm {a = t} (i′ ⊗ k′)) (sum-cong r)

  blocked-ok {s = s ⊗ p} {ι m} {ι n} a b (i′ ⊗ j′ ⊗ k′)
    = blocked-ok (blockₕ a ∘ (i′ ⊗_)) b (j′ ⊗ k′)

  blocked-ok {s = s ⊗ p} {ι m} {q ⊗ r} a b (i′ ⊗ j′ ⊗ (k′ ⊗ l′))
    = blocked-ok (blockₕ a ∘ (i′ ⊗_)) (blockᵥ b ∘ (k′ ⊗_)) (j′ ⊗ l′)

  blocked-ok {s = s ⊗ p} {q ⊗ r} {ι n} a b (i′ ⊗ j′ ⊗ k′)
    = let
        t k = mm (block a ∘ (i′ ⊗ k ⊗_)) (blockₕ b ∘ (k ⊗_))
        r k = blocked-ok (block a ∘ (i′ ⊗ k ⊗_)) (blockₕ b ∘ (k ⊗_)) (j′ ⊗ k′)
      in trans (sum-comm {a = t} (j′ ⊗ k′)) (sum-cong r)

  blocked-ok {s = s ⊗ p} {q ⊗ r} {u ⊗ w} a b (i′ ⊗ j′ ⊗ (k′ ⊗ l′))
    = let
        t k = mm (block a ∘ (i′ ⊗ k ⊗_)) (block b ∘ (k ⊗ k′ ⊗_))
        r k = blocked-ok (block a ∘ (i′ ⊗ k ⊗_))
                         (block b ∘ (k ⊗ k′ ⊗_)) (j′ ⊗ l′)
      in trans (sum-comm {a = t} (j′ ⊗ l′)) (sum-cong r)
\end{code}

\paragraph{Blocking Factors}  The blocking behaviour of \AF{mm}
is fully determined by the shape structure of \AB{s}, \AB{p} and
\AB{q}.  These sub-shapes encode the blocking factors, forcing
us to use arrays of ranks that are larger than two.  We can always
obtain such arrays from regular matrices by choosing multipliers
for rows and columns and organising these multipliers into sub-shapes.
For example, the (rank 4) matrix of shape
(\AC{ι} \AB{m} \AC{⊗} \AC{ι} \AB{n}) \AC{⊗} (\AC{ι} \AB{k} \AC{⊗} \AC{ι} \AB{l})
is isomorphic to the (rank 2) matrix of shape 
\AC{ι} (\AB{m} \AF{*} \AB{n}) \AC{⊗} \AC{ι} (\AB{k} \AF{*} \AB{l}).
We can define conversion functions \AF{flatten-2d} and \AF{unflatten-2d}
that make it possible to switch between the matrices.  

\begin{code}[hide]
module MMreshape 
   (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z)
   (ε-lneutral : ∀ x → ε ⊞ x ≡ x)
   (⊞-assoc : ∀ x y z → x ⊞ (y ⊞ z) ≡ (x ⊞ y) ⊞ z)
   (let infixl 10 _⊠_; _⊠_ = _⊠_) where

  open MMnaive _⊠_ ε _⊞_
\end{code}
\begin{code}
  flatten : S → ℕ
  flatten (ι n)    = n
  flatten (s ⊗ p)  = flatten s * flatten p

  ιflatten : S → S
  ιflatten s = ι (flatten s)

  pflatten : P s → P (ιflatten s)
  punflatten : P (ιflatten s) → P s
\end{code}
The \AF{flatten} function computes the product of all the
leaves in the shape tree, performing multiplications in the order
that is determined by the tree structure.  Index translations \AF{pflatten} and
\AF{punflatten} perform multiplication or division with
remainder (bodies of these functions are omitted).

\begin{code}[hide]
  pflatten {s = ι n} ix = ix
  pflatten {s = p ⊗ q} (ix ⊗ jx) with pflatten ix | pflatten jx
  ... | ι i | ι j = ι (combine i j)

  punflatten {s = ι n} ix = ix
  punflatten {s = s ⊗ p} (ι i)
    = let i , j = quotRem {flatten s} (flatten p) i 
      in punflatten (ι j) ⊗ punflatten (ι i)

  module _ {X : Set} where
\end{code}
Array flattenings simply apply index translations contravariantly.
\begin{code}
    flatten-2d : X ⟪ s ⊗ p ⟫ → X ⟪ ιflatten s ⊗ ιflatten p ⟫
    flatten-2d a (i ⊗ j) = a (punflatten i ⊗ punflatten j)

    unflatten-2d : X ⟪ ιflatten s ⊗ ιflatten p ⟫ → X ⟪ s ⊗ p ⟫
    unflatten-2d a (i ⊗ j) = a (pflatten i ⊗ pflatten j)
\end{code}
With these functions we can turn any matrix of rank 2, into
the matrix of a higher rank by choosing blocking factors for its
rows and columns.  Then we can obtain blocking behaviour by
applying \AF{mm} on the matrix of a higher rank.  We know from
\AF{block-ok} that we will obtain the same elements as if we
were applying \AF{mm-canon} that does not do any blocking, but
it uses all the blocking factors at summation.

If we want to guarantee that blocked algorithm computes
the same result as the canonical one applied to the rank 2
matrix, we have to guarantee that \AF{\_⊞\_} is associative
and \AF{ε} is the neutral element.  Here is the key theorem
that captures this fact.  We omit the proof for space purposes,
but it can be found in~\cite{paper-src}.
\begin{code}[hide]

  lift-ι : ∀ {m n} → (Fin m → Fin n) → P (ι m) → P (ι n)
  lift-ι f (ι i) = ι (f i)


  sum₁-split : ∀ {m n} → (a : Z ⟪ ι (m + n) ⟫)
             → sum₁ _⊞_ ε a 
               ≡ sum₁ {n = m} _⊞_ ε (λ i → a (lift-ι (Fin.inject+ n) i))
                 ⊞ sum₁ {n = n} _⊞_ ε (λ i → a (lift-ι (Fin.raise m) i))
  sum₁-split {m = zero}{n}  a = sym (trans (ε-lneutral _) (sum₁-cong {n = n} λ { (ι i) → refl}))
  sum₁-split {m = suc m}{n} a rewrite sum₁-split {m = m} (a ∘ ιsuc) 
    = trans (⊞-assoc _ _ _) (cong₂ _⊞_ (cong₂ _⊞_ refl (sum₁-cong {n = m} λ { (ι i) → refl}))
                                       (sum₁-cong {n = n} λ { (ι i) → refl}))

 -- sym (trans (sym $ ⊞-assoc _ _ _)
 --                                       (cong₂ _⊞_ refl ?))

--  prod-prod : ∀ {m n} → (a : Z ⟪ ι m ⊗ ι n ⟫)
--            → sum _⊞_ ε a ≡ sum _⊞_ ε (λ i → a (punflatten i))
--  prod-prod {m = zero} a = refl
--  prod-prod {m = suc m}{n} a 
--    = sym (trans (sum₁-split {m = n}{m * n} (λ i → a (punflatten i)))
--                 (cong₂ _⊞_ (sum₁-cong {n = n} first) -- (λ {(ι i) → first {i}}))
--                            (sym (trans (prod-prod {m = m}{n} cutrow {-(λ {(i ⊗ j) → a (ιsuc i ⊗ j)})-})
--                                 (sum₁-cong {n = m * n} second)))))
--      where
--
--        first : (i : P (ι n)) → a (punflatten (lift-ι (Fin.inject+ (m * n)) i)) ≡ a (ι zero ⊗ i)
--        first (ι i) rewrite Fin.splitAt-inject+ n (m * n) i = refl
--
--        cutrow : Z ⟪ ι m ⊗ ι n ⟫
--        cutrow (i ⊗ j) = a (ιsuc i ⊗ j)
--
--        second : ∀ i → (cutrow (punflatten i)) ≡ a (punflatten (lift-ι (Fin.raise n) i))
--        second (ι i) rewrite Fin.splitAt-raise n (m * n) i = refl

  sum₁-nest : ∀ {m n} → (a : Z ⟪ ι (m * n) ⟫)
       → sum₁ _⊞_ ε a
         ≡ sum₁ {n = m} _⊞_ ε (λ {(ι i) → sum₁ _⊞_ ε λ { (ι j) → a (ι (combine i j))}})
  sum₁-nest {m = zero}  a = refl
  sum₁-nest {m = suc m}{n} a rewrite sum₁-split {m = n}{m * n} a 
    = cong₂ _⊞_ 
            (sum₁-cong {n = n} λ { (ι i) → refl})
            (trans (sum₁-nest {m = m} (λ i → a (lift-ι (Fin.raise n) i)))
                   (sum₁-cong {n = m} λ { (ι i) → sum₁-cong {n = n} λ { (ι j) → refl}}))


  quotRem-combine : ∀ {n k} (x : Fin n) y → quotRem {n} k (combine x y) ≡ (y , x)
  quotRem-combine {n}{k} x y = cong swap $ remQuot-combine x y

  flat-sum : (a : Z ⟪ s ⟫)
      → sum _⊞_ ε a ≡ sum _⊞_ ε (λ i → a (punflatten i))
  flat-sum {s = ι n} a = refl
  flat-sum {s = s ⊗ p} a 
    rewrite  flat-sum {s = s} (λ i → sum _⊞_ ε (λ j → a (i ⊗ j))) 
           | sum₁-nest {flatten s}{flatten p} (λ i → a (punflatten i))
    = sum₁-cong {n = flatten s} λ { (ι i) → trans (flat-sum {s = p} (λ j → a (punflatten (ι i) ⊗ j)))
                                                  (sum₁-cong {n = flatten p} λ { (ι j) → helper i j})}
    where
      helper : (i : Fin (flatten s)) (j : Fin (flatten p)) →
               ((λ i₁ → a (punflatten (ι i) ⊗ punflatten i₁)) [ ι j ]) ≡
               ((λ { (ι j) → a (punflatten (ι (combine i j))) }) [ ι j ])
      helper i j rewrite quotRem-combine i j = refl 

\end{code}

\begin{code}
  -- Assuming that ⊞ is associative, and ε is ⊞-neutral
  mm-flat-ok : (a : X ⟪ s ⊗ p ⟫) (b : Y ⟪ p ⊗ q ⟫)
             → mm-canon (flatten-2d a) (flatten-2d b)
               ≈ₐ flatten-2d (mm-canon a b)
\end{code}
\begin{code}[hide]
  mm-flat-ok a b (i ⊗ j) = sym (flat-sum (λ k → a (punflatten i ⊗ k) ⊠ b (k ⊗ punflatten j)))
\end{code}


\paragraph{Pre-blocking}
We can ask ourselves whether it is possible to separate the
blocking steps from the actual recursive matrix multiplication.
That is, would it be possible to pre-block the array, apply
matrix multiplication recursively, and unblock the entire array
back.  To do so, we need to construct an evidence that the
arrays are blocked in a such a way so that all the recursive
calls of \AF{mm} become applicable.  In other words, we have
to capture recursive unfolding of the \AF{mm} for the given
arrays.

To illustrate this, we define a relation that captures one
particular blocking style --- the last case of \AF{mm},
where blocking happens on both the rows and the columns.
Additionally, we chose the outer shape to be always
of the form (\AF{ι} \AB{m} \AC{⊗} \AF{ι} \AB{n}) to prevent
further decomposition on the outer side.  We capture
these assumptions in the inductive relation called \AD{R}.
\begin{code}[hide]
module Preblocked (_⊠_ : X → Y → Z) (ε : Z) (_⊞_ : Z → Z → Z) where

  open MMnaive _⊠_ ε _⊞_
  open MMblocked _⊠_ ε _⊞_
\end{code}
\begin{code}
  data R : S → S → S → Set where
    ι  : R (ι m ⊗ ι n) (ι n ⊗ ι k) (ι m ⊗ ι k)
    σ  : R s p q
       → R ((ι m ⊗ ι n) ⊗ s) ((ι n ⊗ ι k) ⊗ p) ((ι m ⊗ ι k) ⊗ q)
\end{code}
This relation describes all the valid shapes of the two
input matrices and the output matrix, in case they were
pre-blocked in the way we have described above.  The base
case \AC{ι} says that any singletons shapes of the form $m\times n$
$n\times k$ and $n\times k$ satisfy our assumption.
The \AC{σ} case says that if $s$, $p$ and $q$ are
related by \AD{R}, then we can use matrices of the
shape $m\times n$ $n\times k$ and $n\times k$, where
the elements are of the shape $s$, $p$ and $q$
correspondingly.

In this case, if we have a witness for the inner structure
of the argument shapes as described above, we can define
the blocked matrix multiplication inductively as follows:
\begin{code}
  mmx : R s p q → X ⟪ s ⟫ → Y ⟪ p ⟫ → Z ⟪ q ⟫
  mmx ι      a b = mm-canon a b
  mmx (σ r)  a b = let
    a′ = nest a
    b′ = nest b
    c′ = λ { (i ⊗ j) → sum (zipWith _⊞_) (K ε)
                           λ k′ → mmx r (a′ [ i ⊗ k′ ])
                                        (b′ [ k′ ⊗ j ]) }
    in unnest c′
\end{code}
Essentially, we assume that arrays $a$ and $b$ are pre-blocked
(witnessed by the first argument of \AF{mmx}) such that
we always take the last case of \AF{mm} in the non-pre-blocked
scenario.


\paragraph{Pre-blocking correctness}
In order to see that the pre-blocked version \AF{mmx} and
\AF{mm} compute the same results, we have to define conversion
functions  between the original arrays and the pre-blocked ones.

For the given shapes $s$, $p$, and $q$ that are related by \AD{R}
(\ie{} for pre-blocked arrays), we can compute the product shapes
$s′$, $p′$ and $q′$ that will be suitable for \AF{mm}.
Also, we compute the conversion functions from $s$ to $s′ ⊗ p′$,
from $p$ to $p′ ⊗ q′$, and from $s′ ⊗ q′$ into $q$.
\begin{code}
  toinv : ∀ s p q → R s p q
       → Σ[ (s′ , p′ , q′) ∈ S × S × S ]
          (  (∀ {X} → X ⟪ s ⟫ → X ⟪ s′ ⊗ p′ ⟫)
           × (∀ {X} → X ⟪ p ⟫ → X ⟪ p′ ⊗ q′ ⟫)
           × (∀ {X} → X ⟪ s′ ⊗ q′ ⟫ →  X ⟪ q ⟫))
  toinv (ι m ⊗ ι n) (ι n ⊗ ι k) (ι m ⊗ ι k) ι =
    (ι m , ι n , ι k) , id , id , id
  toinv (ι m ⊗ ι n ⊗ s) (ι n ⊗ ι k ⊗ p) (ι m ⊗ ι k ⊗ q) (σ r) = let
    (s′ , p′ , q′) , s⇒s′p′ , p⇒p′q′ , s′q′⇒q = toinv s p q r
    in (ι m ⊗ s′ , ι n ⊗ p′ , ι k ⊗ q′)
     , (λ a → unblock $ unnest $ map s⇒s′p′ $ nest a)
     , (λ a → unblock $ unnest $ map p⇒p′q′ $ nest a)
     , (λ a → unnest $ map s′q′⇒q $ nest $ block a)
\end{code}
\begin{code}[hide]
  postulate
\end{code}

With the given conversions available, we can show that
switching between the pre-blocked version and the blocked one
does not affect the computation.
\begin{code}
    mmx≈mm : (r : R s p q) (a : X ⟪ s ⟫) (b : Y ⟪ p ⟫)
           → let _ , fromA , fromB , toC = toinv s p q r
             in mmx r a b ≈ₐ toC (mm (fromA a) (fromB b))
\end{code}
%\end{AgdaAlign}
You can find the proof in~\cite{paper-src}.  Technically, we have
to compute the conversions in the other direction, but
these conversions are actually isomorphisms.  In order to show this
formally we would have to extend our framework a little further.
We do this in the proof, but not in the paper.



\begin{code}[hide]
fst : P (s ⊗ p) → P s
fst (i ⊗ _) = i
snd : P (s ⊗ p) → P p
snd (_ ⊗ j) = j
_⊛_ : P s → P p → P (s ⊗ p)
i ⊛ j = i ⊗ j
ψ : P (ι n) → Fin n
ψ (ι i) = i
φ : Fin n → P (ι n)
φ = ι

data Reshape : S → S → Set where
  eq     : Reshape s s
  _∙_    : Reshape p q → Reshape s p → Reshape s q
  _⊕_    : Reshape s p → Reshape q r → Reshape (s ⊗ q) (p ⊗ r)
  split  : Reshape (ι (m * n)) (ι m ⊗ ι n)
  flat   : Reshape (ι m ⊗ ι n) (ι (m * n))
  rblock   : Reshape ((s ⊗ p) ⊗ (q ⊗ r)) ((s ⊗ q) ⊗ (p ⊗ r))

_⟨_⟩ : ∀ {s p} → P p → Reshape s p → P s
i ⟨ eq ⟩ = i
i ⟨ r ⊕ r₁ ⟩ = (fst i ⟨ r ⟩) ⊛ (snd i ⟨ r₁ ⟩)
i ⟨ r ∙ r₁ ⟩ = i ⟨ r ⟩ ⟨ r₁ ⟩
i ⟨ split ⟩ = φ $ combine (ψ $ fst i) (ψ $ snd i)
i ⟨ flat ⟩ = let a , b = remQuot _ (ψ i) in φ a ⊛ φ b
(i ⊗ j ⊗ (k ⊗ l)) ⟨ rblock ⟩ = i ⊗ k ⊗ (j ⊗ l)

reshape : Reshape s p → X ⟪ s ⟫ →  X ⟪ p ⟫
reshape r a = λ ix → a [ ix ⟨ r ⟩ ]

rev : Reshape s p → Reshape p s
rev eq = eq
rev (r ⊕ r₁) = rev r ⊕ rev r₁
rev (r ∙ r₁) = rev r₁ ∙ rev r
rev split = flat
rev flat = split
rev rblock = rblock

rblock∘rblock : (i : P (s ⊗ p ⊗ (r ⊗ q))) → i ⟨ rblock ⟩ ⟨ rblock ⟩ ≡ i
rblock∘rblock (i ⊗ j ⊗ (k ⊗ l)) = refl

iota : (n : ℕ) → ℕ ⟪ ι n ⟫
iota zero (ι ())
iota (suc n) (ι zero) = 0
iota (suc n) (ι (suc i)) = map (1 +_) (iota n) (ι i)


r-left : Reshape (ι 8) (ι 2 ⊗ (ι 2 ⊗ ι 2))
r-left = (eq ⊕ split) ∙ split

r-right : Reshape (ι 8) ((ι 2 ⊗ ι 2) ⊗ ι 2)
r-right = (split ⊕ eq) ∙ split

m₂ : ℕ ⟪ (ι 2 ⊗ ι 2) ⊗ (ι 2 ⊗ ι 2) ⟫
m₂ = reshape ((split ⊕ split) ∙ split) (iota 16)

m₃ : ℕ ⟪ (ι 2 ⊗ ι 2) ⊗ (ι 2 ⊗ ι 2) ⟫
m₃ = reshape ((split ⊕ split) ∙ split) (iota 16)


m₂×m₃ : _ --ℕ ⟪ ι 2 ⊗ ι 2 ⟫
m₂×m₃ = MMblocked.mm _*_ 0 _+_ m₂ m₃



prod : S → ℕ
prod (ι x) = x
prod (s ⊗ p) = prod s * prod p

flatten : X ⟪ s ⟫ → X ⟪ ι (prod s) ⟫
flatten {s = ι x} a = a
flatten {s = s ⊗ p} a (ι i) with Fin.remQuot {n = prod s} (prod p) i
... | i , j = flatten (map flatten $ nest a) (ι i) (ι j)

open import Data.List hiding (map; _++_; sum)
toList : X ⟪ ι n ⟫ → List X
toList {n = zero}  x = []
toList {n = suc n} x = x (ι zero) ∷ toList (x ∘ ιsuc)

qq = toList $ flatten m₂×m₃
tt = toList $ flatten $ unblock $ mmx _*_ 0 _+_ (σ ι) (block m₂) (block m₃)
   where open Preblocked

-- 28 ∷ 34 ∷ 76 ∷ 98

open import Data.String hiding (show)
open import Data.Nat.Show

st₁ : String ⟪ ι 2 ⊗ (ι 3 ⊗ ι 4) ⟫
st₁ = reshape ((eq ⊕ split) ∙ split) $ map show (iota 24)

st₂ : String ⟪ (ι 2 ⊗ ι 3) ⊗ ι 4 ⟫
st₂ = reshape ((split ⊕ eq) ∙ split) $ map show (iota 24)


_+ₛ_ : String → String → String
a +ₛ b = "(" ++ a ++ b ++ ")"


st₁+ = sum _+ₛ_ "ε" st₁
st₂+ = sum _+ₛ_ "ε" st₂
-- "(((0(1(2(3ε))))((4(5(6(7ε))))((8(9(10(11ε))))ε)))(((12(13(14(15ε))))((16(17(18(19ε))))((20(21(22(23ε))))ε)))ε))"
-- "(((0(1(2(3ε))))((4(5(6(7ε))))((8(9(10(11ε))))ε)))(((12(13(14(15ε))))((16(17(18(19ε))))((20(21(22(23ε))))ε)))ε))"


\end{code}
